//===- Hello.cpp - Example code from "Writing an LLVM Pass" ---------------===//
//
//                     The LLVM Compiler Infrastructure
//
// This file is distributed under the University of Illinois Open Source
// License. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//
//
// This file implements two versions of the LLVM "Hello World" pass described
// in docs/WritingAnLLVMPass.html
//
//===----------------------------------------------------------------------===//

#define DEBUG_TYPE "DFG"
#include "llvm/Pass.h"
#include "llvm/Function.h"
#include "llvm/Instructions.h"
#include "llvm/BasicBlock.h"
#include "llvm/Constants.h"
#include "llvm/Analysis/MemoryDependenceAnalysis.h"
#include "llvm/Analysis/PostDominators.h"
#include "llvm/Support/CFG.h"
#include "llvm/Type.h"
#include "llvm/ADT/SmallVector.h"
#include "llvm/Support/raw_ostream.h"
#include "llvm/Support/InstIterator.h"
#include "llvm/ADT/Statistic.h"
#include <string>
#include <map>
#include <iostream>
#include <set>
#include <queue>
#include <sstream>
#include <fstream>

using namespace llvm;

STATISTIC(HelloCounter, "Counts number of functions greeted");

namespace {
	class Block{
		public:
		Block(std::string n, int id) { name = n; myID = id; };
		std::string name;
		int myID;
		std::vector<std::string> pred; // names of predecessor and successor blocks in CFG
		std::vector<std::string> succ;
		std::vector<std::string> nodeKeys;
		
		void print(){
			errs()<<"BB Name: "<<name<<"\n";
			errs()<<"\t successors : ";
			for(unsigned int j=0;j<succ.size();j++){
				errs()<<succ[j]<<" ";
			}
			errs()<<"\n\t predecessors : ";
			for(unsigned int j=0;j<pred.size();j++){
				errs()<<pred[j]<<" ";
			}
			errs()<<"\n\t nodeKeys : ";
			for(unsigned int j=0;j<nodeKeys.size();j++){
				errs()<<nodeKeys[j]<<",";
			}
			errs()<<"\n";
		}
	};

   class Nodes{
	public:
	Nodes(std::string n, std::string o1,std::string o2,int oc, int num) {name = n; ops[0]=o1; ops[1]=o2; opcode=oc; opConst[0]=false; opConst[1]=false; errs()<<"Creating "<<name<<" \n"; isMemory=false;instr = num; cmpType=0;};
	Nodes(std::string n, std::string o1,std::string o2, bool o1c, bool o2c, int oc, int num) {name = n;  ops[0]=o1; ops[1]=o2; opcode=oc; opConst[0]=o1c; opConst[1]=o2c; errs()<<"Creating "<<name<<" \n"; isMemory=false; instr = num; cmpType=0;};
	Nodes(std::string n, std::string o1,std::string o2,int oc, std::string fn, int num) {name = n;  ops[0]=o1; ops[1]=o2; opcode=oc; functionName=fn; opConst[0]=false; opConst[1]=false; errs()<<"Creating "<<name<<" \n"; isMemory=false;instr = num; cmpType=0;};
	std::string name;
	std::string ops[2];
	bool opConst[2];
	int cmpType;

/*	std::string op1;
	std::string op2;
	bool op1_Const;
	bool op2_Const;*/
	int instr;
	std::string functionName;
//	std::vector<std::string> pred;
	std::vector<std::string> succ;
	bool isMemory; //if you ever LOAD or STORE to this one set it true
	std::vector<std::string> storeInputs;
	std::vector<unsigned int> GEP_params;

	int opcode;
	int bb; //basic block
	std::vector<int> memDep;

	/***/
	std::vector<std::string> previousDependencies;
	std::vector<int> previousDependenciesDelay;

	/***/
	void print(){
		errs()<<"DESTINATION: "<<name<<" INPUT1: "<<ops[0]<<" INPUT2 "<<ops[1]<<" OPCODE# "<<opcode<<" FunctionName "<<functionName<<" cmpType "<<cmpType<<" \n";
		errs()<<"\t successors : ";
		for(unsigned int j=0;j<succ.size();j++){
			errs()<<succ[j]<<" ";
		}
		
		errs()<<"\n\t storeInputs : ";
		for(unsigned int j=0;j<storeInputs.size();j++){
			errs()<<storeInputs[j]<<" ";
		}
		errs()<<"\n\t memory Dependencies : ";
		for(unsigned int j=0;j<memDep.size();j++){
			errs()<<memDep[j]<<" ";
		}
		
		errs()<<"\n\t GEP params : ";
		for(unsigned int j=0;j<GEP_params.size();j++){
			errs()<<GEP_params[j]<<" ";
		}
		errs()<<"\n";
	}
   };


  class Stores {
	public:
	Stores(std::string dest, std::string block, std::string source, int c, bool constant){
	destination = dest; blockName = block; sourceVar = source; instruction =c; constIn = constant;
	//errs()<<"dest = "<<destination<<" source "<<sourceVar<<"\n";
	};
	int instruction;//instruction number
	std::string destination;//memory location
	std::string blockName;//block in CFG
	std::string sourceVar;//this is usually a register name in our cases
	bool constIn;

	void print(){
		errs()<<"DESTINATION: "<<destination<<" Block: "<<blockName<<" Input: "<<sourceVar<< " Instruction "<<instruction<<" Constant "<<constIn<<"\n"; 
	}
  };
	std::map<std::string,Nodes> theGraph;//DFG
	std::map<std::string,Nodes>::iterator mapIT;
	std::map<std::string,Block> myCFG;
	std::map<std::string,Block>::iterator CFGit;

	std::map<Instruction*,int> instrPointers;
	std::map<Instruction*,int>::iterator ipIT;

	std::map<std::string,int> finishTimes;
	std::map<std::string,int> startTimes;
	std::map<std::string,int>::iterator timesIT;
	std::map<int,int> latencies; //store instruction latencies
	

	std::vector<Stores> allStores;
	std::vector<Stores>::iterator itStores;

	//This convert to array for each BB
/*	typedef	std::map<std::string,int> loadTimes; //
	std::map<std::string,loadTimes> loadTimesList;

	std::map<std::string,loadTimes>::iterator loadTimesListIT;
	loadTimes::iterator loadTimesIT;
*/
	//for generating dependency graph, I need all the loads up to a BB
	typedef std::set<std::string> stringSet;
	typedef	std::map<std::string,stringSet> loadOps; //
	std::map<std::string,loadOps> loadOpsList;

	std::map<std::string,loadOps>::iterator loadOpsListIT;
	loadOps::iterator loadOpsIT;
	//

	std::map<std::string, int> memoryMap;
	std::map<std::string, int>::iterator memoryMapIT;

	std::map<std::string, std::string> GEP_maps;//used to map from addresses created by GEP instruction to real location
	//arrayidx123 -> y[4]
	std::map<std::string, std::string>::iterator GEP_mapsIT;

	std::map<std::string, std::string> BranchMaps;//BB name->cmp where it depends 
	std::map<std::string, std::string>::iterator BranchMapsIT;//BB name->cmp where it depends 
	
	std::map<std::string, int> depths; //depths of nodes in CFG
	
	//Data structures for memory locations renamed in branches
	typedef std::map<std::string,std::string> memoryTemp;
	std::map<std::string, memoryTemp> BlockTempMemory;

	memoryTemp::iterator memoryTempIT;
	std::map<std::string, memoryTemp>::iterator BlockTempMemoryIT;
	
	std::map<std::string, stringSet> PDomMemoryLoc;
	std::map<std::string, stringSet>::iterator PDomMemoryLocIT;

	//End

	//Mapping of instruction types to indexes (index is implied)
	std::vector<int> SupportedOpTypes;
	std::map<int, int> MapOpToALU;
	std::map<int, int>::iterator MapOpToALUit;
//	int WARDelay = 0;
	//

	MemoryDependenceAnalysis* MD;
	AliasAnalysis *AA;
	PostDominatorTree *PDTREE;

	std::vector<Nodes> vectGraph;
	
  // Hello - The first implementation, without getAnalysisUsage.
  struct DFG : public FunctionPass {
    static char ID; // Pass identification, replacement for typeid
    DFG() : FunctionPass(ID) {}

	bool checkOpCode(int opcode){
		return (opcode==Instruction::FAdd || opcode==Instruction::FSub || opcode==Instruction::FMul || opcode==Instruction::FDiv || opcode==Instruction::Call || opcode==Instruction::ICmp || opcode==Instruction::FCmp);
	}


	int StringToInt(std::string s){
		std::stringstream ss;
		ss<<s;
		int number;
		ss>>number;
		return number;
	}

	void ParseConfigFile(std::string filename){
		std::fstream filestr;

		filestr.open (filename.c_str(), std::fstream::in);
		std::string line;
		std::string num;
		if(filestr.fail()){
			errs()<<"Could not open file"<<"\n";
		}

		getline(filestr,line);
		while(!filestr.eof()){
			
			//if(filestr.eof()) break;
			errs()<<line<<"\n";
			if(!(line.find_first_of("#")!=std::string::npos || line.length()==0)){//skip comments & empty lines
				//non comment line
				size_t pos = line.find_first_of(",");//we only want to parse the first number ->(NNN),x,x,x,x,x,x
				if(pos==std::string::npos){
					errs()<<"Error parsing config file \n";
				} else {
					num = line.substr(0, pos);
					int opcode = StringToInt(num);
					num = line.substr(pos+1, line.length() - pos - 1);
					int aluindex = StringToInt(num);
/*					if(number==999){//assign WARDelay
						num = line.substr(pos+2,line.length()-(pos+2));
						WARDelay = StringToInt(num);
					}else {*/
					MapOpToALU.insert(std::pair<int,int>(opcode,aluindex));
//					SupportedOpTypes.push_back(number);
				}
			}
			getline(filestr,line);
		}
		  // >> i/o operations here <<

		filestr.close();

	}

	int OperationTypes (int opcode){
//SupportedOpTypes
		/*switch(opcode){
			case Instruction::FAdd:
				return 0;
			case Instruction::FSub:
				return 1;
			case Instruction::FMul:
				return 2;
			case Instruction::FDiv:
				return 3;
			case Instruction::Call:
				return 4;
			case Instruction::FCmp:
				return 5;
			case Instruction::ICmp:
				return 6;
			default:
				return -1;
		}*/
		MapOpToALUit = MapOpToALU.find(opcode);
		if(MapOpToALUit!=MapOpToALU.end()){
			return MapOpToALUit->second;
		}
/*		for(unsigned int j=0;j<SupportedOpTypes.size();j++){
			if(SupportedOpTypes[j]==opcode)
				return j;
		}*/
		return -1;
	}

/**
Used to search for inputs to a store instruction

**/
	std::string storeInputbyLine(int line, std::vector<Stores> stores){
		for(unsigned int j=0;j<stores.size();j++){
			if(stores[j].instruction==line){
				return stores[j].sourceVar;
			}
		}
		return "";
	}
/**
Find store whose instruction number is 
*/
	std::string storeInputbyDestination(std::string dest, std::vector<Stores> stores){
		for(unsigned int j=0;j<stores.size();j++){
			if(stores[j].destination==dest){
				return stores[j].sourceVar;
			}
		}
		return "";
	}

/**
Return true if the location passed as a parameter is destination of at least 1 store
*/
	bool isStoreLocation(std::string dest, std::vector<Stores> stores){
		for(unsigned int j=0;j<stores.size();j++){
			if(stores[j].destination==dest){
				return true;
			}
		}
		return false;
	}
	/**
Return true if the location passed as a parameter is destination of at least 1 store
It also considers the case of memory locations aliased by GEP instruction, 
such as
arrayidx1 = y[0]
arrayidx2 = y[0]  
these 2 are the same location
*/

	bool isStoreLocationGEP(std::string dest, std::vector<Stores> stores,std::map<std::string,std::string> GEP_map){
		std::string myAlias, theirAlias;
		bool alias = FindNReturn(myAlias,dest,GEP_map);
		bool alias2;
		
		for(unsigned int j=0;j<stores.size();j++){
			if(stores[j].destination==dest && dest!=""){
				return true;
			}
			if(alias){
				alias2 = FindNReturn(theirAlias,dest,GEP_map);
				if(alias2){
					if(myAlias==theirAlias)
						return true;
				}
			}
		}
		return false;
	}
/**
Return the memory location where a source (register) is stored to, or empty string if there isn't any.
Used since we don't propagate store input registers to their store location (since there can be many to one store)
*/
	std::string storeDestByInput(std::string input, std::vector<Stores> stores){
		if(input=="") return input;
		
		for(unsigned int j=0;j<stores.size();j++){
			if(stores[j].sourceVar==input){
				return stores[j].destination;
			}
		}
		return "";
	}

/*
IF a constant value is stored in the destination location of a store, return the constant value.
We return the constant value stored in a memory location, only if that operation is the last store before reaching the instruction number ('instr').
Otherwise do not return anything except a blank string.
*/
	std::string storeInputConst(std::string dest, std::vector<Stores> stores, int instr){
		if(dest=="") return dest;
		
		std::string retval = "";
		int last_instr = 0;
		bool last_const = false;
		for(unsigned int j=0;j<stores.size();j++){
			if(stores[j].destination==dest && stores[j].instruction<instr && stores[j].instruction>last_instr){//must be later that the latest instr so far 
				retval = stores[j].sourceVar;
				last_const = stores[j].constIn;
				last_instr = stores[j].instruction;
			}
		}
		if(last_const)
			return retval;

		return "";
	}
/**

*/

	int cycleOfLatestStore(std::string variable, std::vector<Stores> stores, int instNum,std::map<std::string,int> cycleMap){
		std::map<std::string,int>::iterator mapIT;
		int latestCycle = -1;

		for(unsigned int j=0;j<stores.size();j++){
			if(stores[j].destination==variable && stores[j].instruction<instNum){
				std::string storeinput = stores[j].sourceVar; //look up in map
				mapIT = cycleMap.find(storeinput);
				if(mapIT!=cycleMap.end()){
					latestCycle = (mapIT->second>latestCycle) ? mapIT->second : latestCycle;
				}				
			}
		}
		return latestCycle;
	}

	/*
	return type of operation after getting destination as input
	the destination is used as a key in the map of operations
	*/
	int getOpTypeByDest(std::string dest, std::map<std::string,Nodes> theGraph){
		std::map<std::string,Nodes>::iterator it;
		int optype = -1;
		it = theGraph.find(dest);
		if(it!=theGraph.end()){
			optype = it->second.opcode;
		}
		return optype;
	}
	
	void recordStoreDeps(Nodes * currentNode , std::string variable, std::vector<Stores> stores, int instNum,std::string blockName){
		std::map<std::string,int>::iterator mapIT;
		std::map<std::string,int>::iterator mapIT2;

		for(unsigned int j=0;j<stores.size();j++){
			/** 
				Variables names have to match
				it must occur in an earlier instruction
				Due to the store destination rename that we do for branches
				- if there is a store at the same depth of the CFG 
				- but the block name is different it will not be considered a dependency
					+ because only 1 of them will write to that location when executing.
			**/
			if(stores[j].destination==variable && stores[j].instruction<instNum 
				&& depths[stores[j].blockName]<=depths[blockName] && !(stores[j].blockName!=blockName && depths[stores[j].blockName]==depths[blockName]) ){
				std::string storeinput = stores[j].sourceVar; //look up in map
				//The store input is generated by an operations so it will be in our list
				int optype = getOpTypeByDest(storeinput, theGraph);

//				if(latencies.find(optype)!=latencies.end()){
				if(OperationTypes(optype)!=-1){
					//int delay = latencies[optype];
					int type = OperationTypes(optype);
					currentNode->previousDependencies.push_back(storeinput);
					currentNode->previousDependenciesDelay.push_back(type);
				}
			}
		}
		return ;
	}

	/**
		Return true if a string vector contains an entry or not
	*/
	bool vectorContains(std::vector<std::string> vec, std::string entry){
		std::vector<std::string>::iterator it;
		for(it=vec.begin() ; it < vec.end(); it++ )
			if(*it==entry)
				return true;

		return false;
	}
/**

 Schedule


***/

    void schedule(){
		//BFS exploration of block in CFG
		std::queue<std::string> myqueue;
		std::set<std::string> marked;
		myqueue.push("entry");
		marked.insert("entry");

		while(!myqueue.empty()){
			CFGit = myCFG.find(myqueue.front());
//			errs()<<" PROCESSING "<<myqueue.front()<<"\n";
			myqueue.pop();
			//POP From queue, process it
			Block * currentBlock = &(CFGit->second);
			//Count predecessors of BBlock

			loadOps tempOps;
			loadOps::iterator tempIT;//used to search for entries
			if(currentBlock->pred.size()>=2){
				//merge the previous maps 
				for(unsigned int j=0;j<currentBlock->pred.size();j++){
					if(j==0){
						tempOps = loadOpsList[currentBlock->pred[j]];//get the first map
					}
					else {//for other maps iterate their entries
						for(loadOpsIT= loadOpsList[currentBlock->pred[j]].begin(); loadOpsIT!=loadOpsList[currentBlock->pred[j]].end(); loadOpsIT++){
							//if entry of map exists in first one then merge the sets
							tempIT = tempOps.find(loadOpsIT->first);
							if(tempIT!=tempOps.end()){
								//iterate set entries of the 2nd map
								stringSet::iterator setIT=loadOpsIT->second.begin();
								for(;setIT!=loadOpsIT->second.end();setIT++){
									//insert into current maps set
									//tempIT is the first map
									tempIT->second.insert(*setIT);
								}
							}
						}
					}
				}
				loadOpsList[currentBlock->name] = tempOps;
			}else if(currentBlock->pred.size()==1){
				loadOpsList[currentBlock->name] = loadOpsList[currentBlock->pred[0]];
			}
			
			/** PRINT 
			errs()<<"+"<<currentBlock->name<<"\n";
			for(loadOpsIT= loadOpsList[currentBlock->name].begin(); loadOpsIT!=loadOpsList[currentBlock->name].end(); loadOpsIT++){
				errs()<<loadOpsIT->first<<" ";
				stringSet::iterator setIT;
				for(setIT=loadOpsIT->second.begin();setIT!=loadOpsIT->second.end();setIT++){
					errs()<<*setIT<<" ";
				}
				errs()<<"\n";
			}
			**/
			
			//go through instructions in current block
			std::vector<std::string>::iterator it;
			std::map<std::string,int>::iterator itTimes;
			std::map<std::string,int>::iterator itStartTimes;
			
			for(it = currentBlock->nodeKeys.begin();it != currentBlock->nodeKeys.end(); it++){

				//get opcode, if it is part of those that we can schedule check if the operands are ready
				mapIT = theGraph.find(*it);
				Nodes * currentNode = &(mapIT->second);
				if(!checkOpCode(currentNode->opcode)) // not an operation that we can schedule, usually stores, loads
					continue;

				//If destination of current op is a mem location in allStores, look up the last load time in the table
				//errs()<<currentNode->op1<<","<<currentNode->op2<<"="<<currentNode->name<<"\n";
				std::string storeDest;
				//get finish times of inputs
				for(int idx=0; idx<2; idx++){
					if(!currentNode->opConst[idx] && currentNode->ops[idx]!=""){//if operand is not a constant
						//get finish time

						if(!isStoreLocationGEP(currentNode->ops[idx], allStores,GEP_maps)){//not destination of a store
							int optype = getOpTypeByDest(currentNode->ops[idx], theGraph);
							//if(latencies.find(optype)!=latencies.end()){
							if(OperationTypes(optype)!=-1){
								//record dependencies of inputs, by keeping track of delay and id of operation that generates value
								int type = OperationTypes(optype);
								currentNode->previousDependencies.push_back(currentNode->ops[idx]);
								currentNode->previousDependenciesDelay.push_back(type);
							}
						}

						if(isStoreLocationGEP(currentNode->ops[idx], allStores,GEP_maps)){//if the input is also a location where data has been stored
							
							//Update table of load OPS, since we are going through the inputs of an operation
							//These can only come from loads, if they came from memory (which is checked by isStoreLocation)
							//We only care if it is a store location, because 2 Loads are not conflicts
							loadOpsIT = loadOpsList[currentBlock->name].find(currentNode->ops[idx]);
							if(loadOpsIT==loadOpsList[currentBlock->name].end()){//not inserted
							//	errs()<<"Inserting "<<ops[j]<<"\n";
								std::set<std::string> setTemp;
								setTemp.insert(currentNode->name);
								loadOpsList[currentBlock->name][currentNode->ops[idx]] = setTemp;
								//.insert(std::pair<std::string,(std::set<std::string>)>(currentNode->ops[idx], setTemp));
							} else {//found an entry insert into its set
								loadOpsIT->second.insert(currentNode->name);
							}
							//Deal also with GEP aliasing, if it is aliased we also keep an entry under the name of the alias					
							std::string GEP_str;
							bool alias = FindNReturn(GEP_str,currentNode->ops[idx],GEP_maps);
							if(alias){
								loadOpsIT = loadOpsList[currentBlock->name].find(GEP_str);
								if(loadOpsIT==loadOpsList[currentBlock->name].end()){//not inserted
								//	errs()<<"Inserting "<<ops[j]<<"\n";
									std::set<std::string> setTemp;
									setTemp.insert(currentNode->name);
									loadOpsList[currentBlock->name][GEP_str] = setTemp;
									//.insert(std::pair<std::string,(std::set<std::string>)>(currentNode->ops[idx], setTemp));
								} else {//found an entry insert into its set
									loadOpsIT->second.insert(currentNode->name);
								}
							}
						}
						//If there is a previous store to this location we must wait for it to finish, that is recorded here
						recordStoreDeps(currentNode, currentNode->ops[idx], allStores, currentNode->instr,currentBlock->name);
					}
				}

/* 				for(unsigned int j=0;j<currentNode->memDep.size();j++){
					int dep = currentNode->memDep[j]; // instr number of dependent instruction
					//search for this in Store	graphs data structure and return input operation so we can look up its finish time.
					std::string operation = storeInputbyLine(dep, allStores);
					itTimes = finishTimes.find(operation);
					if(itTimes!=finishTimes.end()){
						maxFinish = ((itTimes->second) > maxFinish) ? (itTimes->second) : maxFinish;
					}
				} */

				//IF we are producing a value that is stored to memory, (if our register is later stored to memory)
				storeDest = storeDestByInput(currentNode->name,allStores);
				//First we find storeDest which is the memory variable name where it is stored, if it is not stored anywhere it will be empty
				// and the next if will fail
				if(isStoreLocationGEP(storeDest, allStores,GEP_maps)){ //if output is a store location => this op is a store
//					int delay = 1;

					/**
						In this case we are storing to memory but there are other loads before us that neeed that location,
						so we wait 1 cycle (delay) before storing to avoid overwriting.
					*/
					loadOpsIT = loadOpsList[currentBlock->name].find(storeDest);
					if(loadOpsIT!=loadOpsList[currentBlock->name].end()){//if there is an OP loading this value wait 1 cycle for them
						stringSet::iterator setIT=loadOpsIT->second.begin();
						for(;setIT!=loadOpsIT->second.end() && *setIT!=currentNode->name;setIT++){ //prevent dependency on itself
							currentNode->previousDependencies.push_back(*setIT);
							currentNode->previousDependenciesDelay.push_back(999);
						}
					}

				}
			/*	if(maxFinish!=0){//only the first operation can start at cycle 0, others must start +1 from previous finishing
					finishTimes[*it]++;
					startTimes[*it]++;
				}*/
			//	errs()<<"currentBlock "<<currentBlock->name<<" operation "<<currentNode->name<<" starttime(maxfinish) "<<maxFinish<<" \n";
			}
			//END of processing
			for(it = currentBlock->succ.begin();it != currentBlock->succ.end(); it++){
				if(marked.count(*it)==0){
					marked.insert(*it);
					myqueue.push(*it);
				}
			}
			//then look at successors
		}
     }

	void memoryLocations(){
		/*
		iterate list of instructions, if it is schedulable find inputs and output, 
		Input: 1) memory location 2) constant 3) register
		Output: 1) Store to mem 2) store to reg
		*/
		std::set<std::string> memorySet;
		std::set<std::string>::iterator memorySetIT;

		for(CFGit=myCFG.begin(); CFGit!=myCFG.end(); CFGit++){
//			errs()<<" PROCESSING "<<myqueue.front()<<"\n";
			//POP From queue, process it
			Block * currentBlock = &(CFGit->second);
			//Count predecessors of BBlock


			//go through instructions in current block
			std::vector<std::string>::iterator it;
			std::map<std::string,int>::iterator itTimes;
			std::string cname="const";
			for(it = currentBlock->nodeKeys.begin();it != currentBlock->nodeKeys.end(); it++){

				//get opcode, if it is part of those that we can schedule check if the operands are ready
				mapIT = theGraph.find(*it);
				Nodes * currentNode = &(mapIT->second);
				if(!checkOpCode(currentNode->opcode)) // not an operation that we can schedule
					continue;

				//If destination of current op is a mem location in allStores, look up the last load time in the table
				//errs()<<currentNode->op1<<","<<currentNode->op2<<"="<<currentNode->name<<"\n";
				std::string storeDest;
				std::string GEP_str;
				//get finish times of inputs

				//Input operands
				for(int idx=0; idx<2; idx++){
					if(currentNode->ops[idx]!=""){//If it is empty there is nothing to do
						if(FindNReturn(GEP_str,currentNode->ops[idx],GEP_maps))//first search in GEP mappings (aliases), if it is there insert that value not the alias
							memorySet.insert(GEP_str);
						else 
							memorySet.insert(currentNode->ops[idx]);
						//errs()<<"inserting1 :"<<currentNode->op1<<"\n";					
					}
				}

				/**
					If an instruction produces a value that is stored to memory
					allocate space for the destination memory location.
					tmp = add x,y
					store tmp, mem1  // we need to create a location for mem1
					if
					tmp = add x,y
									//we need to create a location for tmp
									
					Since we use sets there is no need to worry about duplicates
				*/
				storeDest = storeDestByInput(currentNode->name,allStores);
				if(storeDest!=""){//if this is a store use the store address
					if(FindNReturn(GEP_str,storeDest,GEP_maps))//check if aliased
						memorySet.insert(GEP_str);
					else 
						memorySet.insert(storeDest);
					//errs()<<"inserting :"<<storeDest<<"\n";
				}
				else {//if the value is not stored still create a temporary location
					if(FindNReturn(GEP_str,currentNode->name,GEP_maps))
						memorySet.insert(GEP_str);
					else 
						memorySet.insert(currentNode->name);
					//errs()<<"inserting :"<<currentNode->name<<"\n";
				}
			}
		}
		
		/*
		temporary solution for stores of constants, insert them in the set of locations
		then deal with the aliasing
		*/
		for(itStores=allStores.begin(); itStores !=allStores.end(); itStores++){
			if(itStores->constIn){
				memorySet.insert(itStores->sourceVar);
			}
		}
		
		//errs()<<"Printing memory locations\n";
		int memIndex= 0;
		for(memorySetIT=memorySet.begin();memorySetIT!=memorySet.end();memorySetIT++){
			errs()<<memIndex<<" <-= "<<*memorySetIT<<"\n";
			memoryMap.insert(std::pair<std::string, int>(*memorySetIT, memIndex++));
		}
	}
	/**
	Find a key in a map, return the key by reference and return true otherwise return false
	*/
	bool FindNReturn(std::string &Value, std::string key, std::map<std::string,std::string> map){
		std::map<std::string,std::string>::iterator mapIT;
		mapIT = map.find(key);
		if(mapIT==map.end())
			return false;

		Value = mapIT->second;
		return true;	
	}
	
	void GEP_mappings(){
		//Creating GEP mappings
		for(mapIT= theGraph.begin();mapIT!=theGraph.end();mapIT++){
			int opcode = mapIT->second.opcode;
			if(opcode==Instruction::GetElementPtr){//create new entry in GEP_maps
				std::string key = mapIT->second.name;
				//create var name of GEP access using input1 and the numerical parameters
				std::string varname = mapIT->second.ops[0];
				std::stringstream ss;
				//GEP_maps
				for(unsigned int j =0 ; j< mapIT->second.GEP_params.size(); j++){
					ss<<"_"<<mapIT->second.GEP_params[j];
				}
				varname += ss.str();
				//errs()<<"Created varname "<<varname<<"\n";
				GEP_maps.insert(std::pair<std::string,std::string>(key,varname));
			}
		}
	}
/**
Output format
destID,opType,input1(OP)input2,startcycle,input1Address,input2Address,outputAddress

*/
	void schedulePrinter(Function &F){
		std::string BBName;
		std::vector<std::string>::iterator it;
		
		
		std::queue<std::string> myqueue;
		std::set<std::string> marked;
		myqueue.push("entry");
		marked.insert("entry");

		while(!myqueue.empty()){
			CFGit = myCFG.find(myqueue.front());
			myqueue.pop();
			//POP From queue, process it
			Block * currentBlock = &(CFGit->second);


			for(it = currentBlock->nodeKeys.begin();it != currentBlock->nodeKeys.end(); it++){
				//Look up name in start times ( if it is schedulable it will have an entry)
				std::string name=*it;//I->getName();
				
				mapIT = theGraph.find(name);
				
				if(checkOpCode(mapIT->second.opcode)){
					
					std::stringstream ss;
					
					int type = OperationTypes(mapIT->second.opcode);
					errs() << name <<","<<type<<",";
			
					bool newmem;
					std::string GEP_str;
					std::string inputVal;
					std::string storeDest;
					std::string storeConst;
					std::string memKey;

					for(int idx=0; idx<2; idx++){
						/*storeDest = storeDestByInput(mapIT->second.ops[idx],allStores);
						memKey=mapIT->second.ops[idx];
						if(storeDest!=""){
						errs()<<"\n storeDest "<<storeDest<<"\n";
						errs()<<"mapIT->second.ops[idx];"<<mapIT->second.ops[idx]<<"\n";
						}
						if(storeDest!=""){//if this is a store use the store address
							inputVal = storeDest;
						}else {
							inputVal = mapIT->second.ops[idx];
						}*/
						/**
							If the value is a destination of a previous store this is Ok unless that store is storing a constant.
							We should always look at the most recent store. The store destination rename we do in branches is useful otherwise it would be more complicated to deal with stores

							In that case I need to find the address of the constant.
							In the first case I still need to check the GEP maps if the location is aliased. 
						*/
						//temporary fix for constants
						inputVal = mapIT->second.ops[idx];
						storeConst = storeInputConst(inputVal,allStores,mapIT->second.instr);
						if(storeConst!=""){
							inputVal = storeConst;
						}
						newmem = FindNReturn(GEP_str,inputVal,GEP_maps);
						//errs()<<"GEP_str "<<GEP_str<<"\n";
						//errs()<<"inputVal "<<inputVal<<"\n";
						memKey = (newmem) ? GEP_str : inputVal;
						errs()<<memKey;
						if(idx==0)
							errs()<<"(OP)";

						memoryMapIT = memoryMap.find(memKey);
						if(memoryMapIT!=memoryMap.end()){
							ss<<memoryMapIT->second;
						}else {
							ss<<"0";//for easier parsing on the other side
						}
						ss<<",";
					}

					/** Find address of destination register */
					storeDest = storeDestByInput(name,allStores);
					if(storeDest!=""){//if this is a store use the store address
						memKey = storeDest;
					}else {
						
						memKey = name;
					}
					newmem = FindNReturn(GEP_str,memKey,GEP_maps);
					memKey = (newmem) ? GEP_str : memKey;

					memoryMapIT = memoryMap.find(memKey);
					if(memoryMapIT!=memoryMap.end()){
							ss<<memoryMapIT->second;
					} // destination address
					
					//errs()<<","<<"timesIT->second"<<",";//start cycle
					errs()<<",";
					errs()<<ss.str()<<","<<mapIT->second.cmpType<<"\n";
					
					for(unsigned int j=0; j<mapIT->second.previousDependencies.size();j++){
						errs()<<mapIT->second.previousDependencies[j]<<" ";
					}
					errs()<<"\n";
					for(unsigned int k=0; k<mapIT->second.previousDependenciesDelay.size();k++){
						errs()<<mapIT->second.previousDependenciesDelay[k]<<" ";
					}
					errs()<<"\n";
				}
			}

			for(it = currentBlock->succ.begin();it != currentBlock->succ.end(); it++){
				if(marked.count(*it)==0){
					marked.insert(*it);
					myqueue.push(*it);
				}
			}
		}
		//for (Function::iterator b = F.begin(), e = F.end(); b != e; ++b){
		/*
		for(CFGit = myCFG.begin(); CFGit != myCFG.end(); CFGit++){

		  //BBName = b->getName();
		  BBName = CFGit->first;
		  Block * currentBlock = &(CFGit->second);
		  //errs()<<"\n BB name = "<<BBName<<" \n"; 
		  //CFGit = myCFG.find(b->getName()); //Look up BB name and get a pointer to the data
			//errs()<<"\n";

			//errs()<<"\tNode"<<BBName <<"\n ";
			
			//for (BasicBlock::iterator I = b->begin(), ie = b->end(); I != ie; ++I) {
		}
	*/
	}

	void printDOT(Function &F){
		for (Function::iterator b = F.begin(), e = F.end(); b != e; ++b){
		  size_t found;
		  std::string BBName = b->getName();
		  //errs()<<"\n BB name = "<<BBName<<" \n"; 
		  CFGit = myCFG.find(b->getName()); //Look up BB name and get a pointer to the data
			errs()<<"\n";
		
			found = BBName.find(".");
			if(found!=std::string::npos){
				BBName = BBName.erase(found,1);
			}
			errs()<<"\tNode"<<BBName <<" [shape=record,label=\"{"<<BBName<<" \\l ";
		
			for (BasicBlock::iterator I = b->begin(), ie = b->end(); I != ie; ++I) {
				//Look up name in start times ( if it is schedulable it will have an entry)
				std::string name=I->getName();
				errs() << *I <<" ";
				timesIT= startTimes.find(name);
				if(timesIT!=startTimes.end()){
					errs()<<"(( "<<timesIT->second <<" ))";
				}
				errs()<< " \\l ";
			}
			//succ
			int succ_size = CFGit->second.succ.size();
			if(succ_size<=1){
				errs()<<"}\"];\n";
			} else {
				errs()<<"|{<s0>T|<s1>F}}\"];\n";
			}
			for(unsigned int j=0;j<CFGit->second.succ.size();j++){
				std::string succ = CFGit->second.succ[j];
				found = succ.find(".");
				if(found!=std::string::npos){
					succ = succ.erase(found,1);
				}
				if(succ_size>1){
					errs()<<"\tNode"<<BBName<<":s"<<j<<" -> "<<"Node"<<succ<<";\n";
				}else {
					errs()<<"\tNode"<<BBName<<" -> "<<"Node"<<succ<<";\n";
				}
			}
		
		}
	}

	void Handle_Alloca(int opcode, int instructionCount, std::string dest){
		Nodes * nPtr=NULL;
		errs()<<"Allocate \n";
/*		for(Instruction::op_iterator opi = (*I).op_begin(); opi != (*I).op_end(); opi++){
//					errs() << opi->get()->getType()->getTypeID() << " ";
			errs() << "\t type= "<<opi->get()->getType()->getTypeID()<< " Name = "<<opi->get()->getName() << "\n";
			source[j++] = opi->get()->getName();
		}*/
		nPtr = new Nodes(dest, "", "",opcode,instructionCount);
		nPtr->isMemory = true;
		theGraph.insert(std::pair<std::string,Nodes>(dest,*nPtr)); //Global Variable
		//finishTimes.insert(std::pair<std::string,int>(dest,-1));// map of finish times
		//Allocate does not cause any data flow
	}
	
	void Handle_Load(Instruction * I, int opcode, int instructionCount,std::string dest, std::string BBName){
		errs()<<"Load \n";
		int j=0;
		std::string source[5];
		Nodes * nPtr=NULL;
		bool cmpdependency=false;
		BranchMapsIT = BranchMaps.find(BBName); //find CMP where(if) we depend
		if(BranchMapsIT!=BranchMaps.end()){
			cmpdependency=true;
		}

		for(Instruction::op_iterator opi = (*I).op_begin(); opi != (*I).op_end(); opi++){
//					errs() << opi->get()->getType()->getTypeID() << " ";

			errs() << "\t type= "<<opi->get()->getType()->getTypeID()<< " Name = "<<opi->get()->getName() << "\n";
			source[j++] = opi->get()->getName();
		}
		if(cmpdependency){
			//search for an entry in the block maps in case a previous store has renamed this location
			BlockTempMemoryIT = BlockTempMemory.find(BBName);
			if(BlockTempMemoryIT!=BlockTempMemory.end()){//search for entry for current block
				memoryTempIT = BlockTempMemory[BBName].find(source[0]);
				if(memoryTempIT!=BlockTempMemory[BBName].end()){//if we don't find an entry for memory location insert a new one, else we just use the same
					source[0] = memoryTempIT->second;
				}
			}
		}
		nPtr = new Nodes(dest, source[0], "",opcode,instructionCount); 
		nPtr->isMemory = true;
		theGraph.insert(std::pair<std::string,Nodes>(dest,*nPtr));//Global variable
		mapIT=theGraph.find(source[0]);
		if(mapIT!=theGraph.end()){
			mapIT->second.succ.push_back(dest); //Y = LOAD X, keep track of data flow from X to Y by making Y successor of X
		}
		//finishTimes.insert(std::pair<std::string,int>(dest,0));// map of finish times
	}
	
	void Handle_Branch(Instruction * I){
		errs()<<"Branch ";
		if(BranchInst *branch = dyn_cast<BranchInst>(I)){
			if(branch->isConditional()){
				errs()<<"CONd branch \n";
				errs()<<branch->getCondition()->getName()<<"\n";
				
				for(unsigned int idx =0; idx<branch->getNumSuccessors();idx++){
					errs()<<"br "<<idx<<" "<<branch->getSuccessor(idx)->getName()<<" "<<branch->getCondition()->getName()<<"\n";
					if(branch->getSuccessor(idx)->getName().find(".end")==std::string::npos){
						BranchMaps.insert(std::pair<std::string,std::string>(branch->getSuccessor(idx)->getName(),branch->getCondition()->getName()));
					}
				}
			}
		
		}
	}
	
	void Handle_CMP(Instruction * I, int opcode, int instructionCount, std::string dest){
		errs()<<"CMP ";
		bool constant[2];
		std::string source[2];
		constant[0] = false;
		constant[1] = false;
		int j = 0;
		Nodes * nPtr=NULL;
		for(Instruction::op_iterator opi = (*I).op_begin(); opi != (*I).op_end(); opi++){//Iterate the operands of the CMP instruction
//					errs() << opi->get()->getType()->getTypeID() << " ";
			errs() << "\n\t type= "<<opi->get()->getType()->getTypeID()<< " Name = "<<opi->get()->getName() << "";
			errs() << "\nGet value ID = "<<opi->get()->getValueID() << "";

			ConstantFP *FP_ptr = dyn_cast<ConstantFP>(opi->get());
			SmallVector<char,0> strfl;
			SmallVectorImpl<char>::iterator flIT;
			
			//Now the operands either have a name which means it is a variable or it is a ConstantFP which is Floating point value 
			
			if( opi->get()->hasName()==true) {
				source[j++] = opi->get()->getName();
				mapIT=theGraph.find(source[j-1]);
				if(mapIT!=theGraph.end()){
					mapIT->second.succ.push_back(dest);//create dependency between current CMP and operand that we are reading
				}
			}
			else if ((FP_ptr) && (opi->get()->hasName()==false)) {//is the constant a FP value
				//The following is a complicated way to convert a constant FP value into a string and put it into an entry of
				//a source array
				(FP_ptr->getValueAPF()).toString(strfl,0,3); //convert to a string
				  errs()<<"\nFloat val: ";
				source[j] = "";
				constant[j] = true;
				for(flIT= strfl.begin(); flIT!= strfl.end(); flIT++){
					errs()<<*flIT;
					source[j].push_back(*flIT);
				}
				j++;
			} else {
				//Some case that I did not handle occurred
				errs()<<"BIG PROBLEMM0 \n\n\n\n\n";
				exit(1);
			}

		}
		
		errs()<< " Full names ";
		nPtr = new Nodes(dest, source[0], source[1],constant[0], constant[1],opcode,instructionCount);
		if(CmpInst *cmp = dyn_cast<CmpInst>(I)){
			nPtr->cmpType = cmp->getPredicate();//Store the predicate type, for this refer to LLVM documentation.
		}					
		theGraph.insert(std::pair<std::string,Nodes>(dest,*nPtr));
		//finishTimes.insert(std::pair<std::string,int>(dest,-1));// map of finish times
	}
	
 
	void Handle_GEP(Instruction * I,int opcode, int instructionCount, std::string dest){
		errs()<<"GetElementPtr \n";
		int j=0;
		std::string source[5];
		Nodes * nPtr=NULL;
		
		for(Instruction::op_iterator opi = (*I).op_begin(); opi != (*I).op_end(); opi++){
//					errs() << opi->get()->getType()->getTypeID() << " ";
			errs() << "\t type= "<<opi->get()->getType()->getTypeID()<< " Name = "<<opi->get()->getName() << "\n";
			if(j==0){//only the first operand is a named value
				source[j++] = opi->get()->getName();
				errs()<<source[j-1]<<"\n";
				mapIT=theGraph.find(source[j-1]);	
				if(mapIT!=theGraph.end()){ //update successors
					//errs()<<mapIT->first;
					mapIT->second.succ.push_back(dest);
				}
				nPtr = new Nodes(dest, source[0], "",opcode,instructionCount); //create the entry in the list of operations
				theGraph.insert(std::pair<std::string,Nodes>(dest,*nPtr));
			}else {
				/* Now after creating the entry, then we get the operands of the GEP instruction which are constant indexes
				that indicate how to calculate the address/pointer. These are stored in the GEP_params structure of the entry
				*/
				if(ConstantInt *I_ptr = dyn_cast<ConstantInt>(opi->get())){
					errs()<<"Constantint ";
					errs()<<I_ptr->getValue()<<"\n";
					mapIT=theGraph.find(dest);	
					if(mapIT!=theGraph.end()){
						unsigned int number = static_cast<unsigned int>(*(I_ptr->getValue().getRawData())); //funny way to get an int out of it
						errs()<<number<<"\n";
						mapIT->second.GEP_params.push_back(number);
					}
				}
			}
		}
		//finishTimes.insert(std::pair<std::string,int>(dest,0));// map of finish times
	}
 
    void Handle_Binary(Instruction * I,int opcode, int instructionCount, std::string dest){
		bool constant[2];
		constant[0] = false;
		constant[1] = false;
		Nodes * nPtr = NULL;
		int j =0;
		std::string source[2];
		
		for(Instruction::op_iterator opi = (*I).op_begin(); opi != (*I).op_end(); opi++){
	//					errs() << opi->get()->getType()->getTypeID() << " ";
			errs() << "\n\t type= "<<opi->get()->getType()->getTypeID()<< " Name = "<<opi->get()->getName() << "";
			errs() << "\nGet value ID = "<<opi->get()->getValueID() << "";

			ConstantFP *FP_ptr = dyn_cast<ConstantFP>(opi->get());
				SmallVector<char,0> strfl;
				SmallVectorImpl<char>::iterator flIT;
			
			/*
			Operand of binary(meaning 2 operand) instruction can be a constant or a variable
			If it has a name is a variable
			*/
			
			if( opi->get()->hasName()==true) {//record name and keep track of dependencies in succ(essor) structure
				source[j++] = opi->get()->getName();
				mapIT=theGraph.find(source[j-1]);	
				if(mapIT!=theGraph.end()){
					//errs()<<mapIT->first;
					mapIT->second.succ.push_back(dest);
				}
			}
			else if ((FP_ptr) && (opi->get()->hasName()==false)) {//is the constant a FP value
				(FP_ptr->getValueAPF()).toString(strfl,0,3); //convert to a string
				errs()<<"\nFloat val: ";
				source[j] = "";
				constant[j] = true;
				for(flIT= strfl.begin(); flIT!= strfl.end(); flIT++){
					errs()<<*flIT;
					source[j].push_back(*flIT);
				}
				j++;
			} else {
				errs()<<"BIG PROBLEMM0 \n\n\n\n\n";
				exit(1);
			}

		}

		errs()<< " Full names ";
		nPtr = new Nodes(dest, source[0], source[1],constant[0], constant[1],opcode,instructionCount);
		theGraph.insert(std::pair<std::string,Nodes>(dest,*nPtr));

		//finishTimes.insert(std::pair<std::string,int>(dest,-1));// map of finish times
	}
	
	void Handle_Call(CallInst* callInst, int opcode, int instructionCount, std::string dest){
		//errs()<<" Call inst \n";
		//Dest is already set,
			//TODO: support more than 2 arguments, support constants as arguments
		std::string source[2];
		
		Nodes * nPtr=NULL;
		errs()<<callInst->getCalledFunction()->getName()<<":";
		for(unsigned int j=0;j<callInst->getNumArgOperands();j++){
			errs()<<callInst->getArgOperand(j)->getName()<<" ";
			source[j] = callInst->getArgOperand(j)->getName();
		}
		errs() << "\n";
		nPtr = new Nodes(dest, source[0], source[1],opcode,callInst->getCalledFunction()->getName(),instructionCount ); 
		theGraph.insert(std::pair<std::string,Nodes>(dest,*nPtr));
		mapIT=theGraph.find(source[0]);
		if(mapIT!=theGraph.end()){
			mapIT->second.succ.push_back(dest);
		}
		//finishTimes.insert(std::pair<std::string,int>(dest,-1));// map of finish times
	}

	void Handle_Stores(Instruction * I,int opcode, int& instructionCount,std::string BBName,BasicBlock * b){
		/*** Special case for branches***/
		/** If this is a store inside a branch, rename the store destination,
			add a new instruction cmp olddest, newdest, type = 0
			add a dependency for CMP to the instruction that produces input of store
		**/
		int j=0;
		std::string source[2];
		std::string cmpName;
		std::string origStore;
		Nodes * nPtr=NULL;
		bool cmpdependency=false;
		BranchMapsIT = BranchMaps.find(CFGit->first); //find CMP where(if) we depend
		if(BranchMapsIT!=BranchMaps.end()){
			cmpName = BranchMapsIT->second;
			errs() << "CMP NAME "<<cmpName <<"\n";
			cmpdependency=true;
		}
	
		errs()<<"Store \n";
		bool constant = false;
		for(Instruction::op_iterator opi = (*I).op_begin(); opi != (*I).op_end(); opi++){
//					errs() << opi->get()->getType()->getTypeID() << " ";

			ConstantFP *FP_ptr = dyn_cast<ConstantFP>(opi->get());
			SmallVector<char,0> strfl;
			SmallVectorImpl<char>::iterator flIT;
			
			errs() << "\t type= "<<opi->get()->getType()->getTypeID()<< " Name = "<<opi->get()->getName() << "\n";
			if(opi->get()->hasName()){
				source[j++] = opi->get()->getName();
			}else if ((FP_ptr) && (opi->get()->hasName()==false)) {//is the constant a FP value
				(FP_ptr->getValueAPF()).toString(strfl,0,3); //convert to a string
				  errs()<<"\nFloat val: ";
				source[j] = "";
				for(flIT= strfl.begin(); flIT!= strfl.end(); flIT++){
					errs()<<*flIT;
					source[j].push_back(*flIT);
				}
				if(j==0) constant = true;
				j++;
			}
		}
		if(cmpdependency){
			origStore = source[1];
			//Insert entry of renamed memory location
			//search for entry , if it not existing insert
			BlockTempMemoryIT = BlockTempMemory.find(BBName);
			if(BlockTempMemoryIT!=BlockTempMemory.end()){//search for entry for current block
				memoryTempIT = BlockTempMemory[BBName].find(origStore);
				if(memoryTempIT==BlockTempMemory[BBName].end()){//if we don't find an entry for memory location insert a new one, else we just use the same
					source[1] = source[1].append(source[0]);//create new destination, this is unique because we used name of source reg
					BlockTempMemory[BBName][origStore] = source[1];
				} else {//mapping was created by previous instruction
					source[1] = BlockTempMemory[BBName][origStore]; //use previous mapping
				}
			} else {
				source[1] = source[1].append(source[0]);//create new destination, this is unique because we used name of source reg
				BlockTempMemory[BBName][origStore] = source[1];
			}
			BasicBlock * NearCDOM = PDTREE->findNearestCommonDominator(b,b->getSinglePredecessor());
			//errs()<<"\n\n\n\n\n=====================\n "<<NearCDOM->getName()<<"\n\n";
			std::string PDOMname = NearCDOM->getName();
			PDomMemoryLocIT = PDomMemoryLoc.find(PDOMname);
			if(PDomMemoryLocIT!=PDomMemoryLoc.end()){
				PDomMemoryLocIT->second.insert(origStore);
			} else {
				stringSet tempset;
				tempset.insert(origStore);
				PDomMemoryLoc.insert(std::pair<std::string,stringSet>(PDOMname,tempset));
			}
		}

		nPtr = new Nodes(source[1], source[0], "",opcode,instructionCount); //Destination of store is second operand
		nPtr->isMemory = true;
		mapIT = theGraph.find(source[1]);
		if(mapIT==theGraph.end()){
			theGraph.insert(std::pair<std::string,Nodes>(source[1],*nPtr));
		}else {
			mapIT->second.storeInputs.push_back(source[0]);
		}
		//vectGraph.push_back(*nPtr);

		mapIT=theGraph.find(source[0]);//what we store
		if(mapIT!=theGraph.end()){
			mapIT->second.succ.push_back(source[1]);//where we store it
		}

		CFGit->second.nodeKeys.push_back(source[1]);
		//finishTimes.insert(std::pair<std::string,int>(source[1],-1));
		
		//** Stores data structure */
		Stores * st = new Stores(source[1], BBName, source[0], instructionCount, constant);
		allStores.push_back(*st);
	}
	
	
	virtual void getAnalysisUsage(AnalysisUsage &AU) const {
	AU.addRequired<MemoryDependenceAnalysis>();
	AU.addRequired<AliasAnalysis>();
	AU.addRequired<PostDominatorTree>();
    }

    virtual bool runOnFunction(Function &F) {
	MD = &getAnalysis<MemoryDependenceAnalysis>();
	AA = &getAnalysis<AliasAnalysis>();	
	PDTREE = &getAnalysis<PostDominatorTree>();	
	//For each BB keep a set of keys to the map of DFG entries
	latencies[Instruction::FAdd] = 7;
	latencies[Instruction::FSub] = 7;
	latencies[Instruction::FMul] = 5;
	latencies[Instruction::FDiv] = 6;
	latencies[Instruction::Call] = 17;
	latencies[Instruction::FCmp] = 1;
	latencies[Instruction::ICmp] = 1;
ParseConfigFile("config.txt");

	Nodes * nPtr=NULL;
	Block * bPtr = NULL;
    ++HelloCounter;
	int opcode = -1;
      errs() << "Hello: ";
      errs().write_escaped(F.getName()) << '\n';
	  int bbNumber = 0;//used for BB ID TODO: maybe remove
	  
	std::string BBName;
	int instructionCount=0;
	for (Function::iterator b = F.begin(), e = F.end(); b != e; ++b){
	  
	  BBName = b->getName();
		errs()<<"BB name = "<<BBName<<" \n";

	  /** Create CFG **/
	  bPtr  = new Block(b->getName(),bbNumber++);
	  myCFG.insert(std::pair<std::string,Block>(static_cast<std::string>(b->getName()),*bPtr));
	  
	  CFGit = myCFG.find(b->getName());
	  
		errs()<<"Pred ";
		for (pred_iterator PI = pred_begin(b), E = pred_end(b); PI != E; ++PI) {
			BasicBlock *Pred = *PI;
			errs()<<Pred->getName()<< " ";
			CFGit->second.pred.push_back(Pred->getName());
		}
		errs()<<"Succ ";
		for (succ_iterator PI = succ_begin(b), E = succ_end(b); PI != E; ++PI) {
			BasicBlock *Succ = *PI;
			errs()<<Succ->getName()<< " ";
			CFGit->second.succ.push_back(Succ->getName());
		}
		/** END : Create CFG **/
		errs()<<"\n";
		/** Create DFG **/
		
		bool done= false;
		bool PDOMcheck =false;
		for (BasicBlock::iterator I = b->begin(), ie = b->end(); I != ie; ++I) {
			//nodeKeys
			bool names = true;
			std::string dest, source[5];
			for(int i=0;i<5;i++)
				source[i]="";
			
			//Instruction Map
			instrPointers.insert(std::pair<Instruction*,int>(I,instructionCount));
			
			errs() << "Destination:" << I->getName() <<" ,"; //getOpCodeName()
			
			dest = I->getName();
			
			errs() << "Opcode:" << I->getOpcode() <<" "; //getOpCodeName()
			opcode = I->getOpcode();


			if(opcode!=Instruction::Store && !I->isTerminator()){
				CFGit->second.nodeKeys.push_back(dest);
			}
			
			/*
			if(done)
				exit(1);
			*/
			
			if(I->hasName()==false){
				names = false;
				done = true;
			}


			int j=0;
			/*
			Handle different types of instructions
			*/
			if(!PDOMcheck){
				//Check postdom entries for an entry for the current block and insert appropriate cmp instructions
				PDomMemoryLocIT = PDomMemoryLoc.find(BBName);
				if(PDomMemoryLocIT!=PDomMemoryLoc.end()){
					errs()<<"\n\nPostdom "<<BBName<<"\n";		
					stringSet::iterator ssIT;
					for(ssIT=PDomMemoryLocIT->second.begin();ssIT!=PDomMemoryLocIT->second.end();ssIT++){
						errs()<<"-"<<*ssIT<<"\n";
						std::string memLoc[2];
						int inuse = -1;
						//For each memory location find its corresponding temporary in the predecessors of the block and 
						//use it to create a cmp instruction
						CFGit = myCFG.find(BBName);
						
						for(unsigned int j=0;j<CFGit->second.pred.size();j++){
							std::string predName =CFGit->second.pred[j];
							memoryTempIT = BlockTempMemory[predName].find(*ssIT);
							if(memoryTempIT!=BlockTempMemory[predName].end()){
								memLoc[j] = memoryTempIT->second;
								inuse = j;
							} else {
								memLoc[j] = *ssIT; // if there is no entry just use the original, it will copy the same value
							}
						}//end for pred
						std::string newname = "cmp_"+memLoc[inuse];//make unique by using source register (SSA)
						nPtr = new Nodes(newname, memLoc[0], memLoc[1],Instruction::FCmp,++instructionCount); //Destination of store is second operand
						nPtr->cmpType = 0;
						for(unsigned int j=0;j<CFGit->second.pred.size();j++){
							std::string predName = CFGit->second.pred[j];
							BranchMapsIT = BranchMaps.find(predName); //find CMP where(if) we depend
							if(BranchMapsIT!=BranchMaps.end()){
								//int delay = latencies[Instruction::FCmp];
								int type = OperationTypes(Instruction::FCmp);
								if(!vectorContains(nPtr->previousDependencies,BranchMapsIT->second))
								{
									nPtr->previousDependencies.push_back(BranchMapsIT->second);
									nPtr->previousDependenciesDelay.push_back(type);
								}
							}
						}//end for pred
						
						theGraph.insert(std::pair<std::string,Nodes>(newname,*nPtr));
						CFGit->second.nodeKeys.push_back(newname);
						Stores * st = new Stores(*ssIT, BBName, newname, instructionCount, false);
						allStores.push_back(*st);
					}//end for
				}//end if found
				PDOMcheck = true;
				
			/*std::string newname = "cmp_"+source[0];//make unique by using source register (SSA)
			nPtr = new Nodes(newname, source[1], origStore,Instruction::FCmp,++instructionCount); //Destination of store is second operand
			nPtr->cmpType = 0;
	
			int delay = latencies[Instruction::FCmp];
			nPtr->previousDependencies.push_back(cmpName);
			nPtr->previousDependenciesDelay.push_back(delay);
			
			theGraph.insert(std::pair<std::string,Nodes>(newname,*nPtr));
			CFGit->second.nodeKeys.push_back(newname);
			//Add new store for the new instruction, cmp_source[0], since this instruction may store to any of the 2 locations
			//we have to keep track of both of them
			Stores * st = new Stores(source[1], BBName, newname, instructionCount, false);
			allStores.push_back(*st);
			st = new Stores(origStore, BBName, newname, instructionCount, false);
			allStores.push_back(*st);*/
			}


			if(opcode==Instruction::Store){
				Handle_Stores(I,opcode, instructionCount,BBName,b);
			}else if(opcode==Instruction::Alloca){
				Handle_Alloca(opcode,instructionCount,dest);
			}
			else if(opcode==Instruction::Load){
				Handle_Load(I, opcode, instructionCount,dest,BBName);
			}else if(opcode==Instruction::Br){
				Handle_Branch(I);
			}else if(opcode==Instruction::FCmp || opcode==Instruction::ICmp){
				Handle_CMP(I, opcode, instructionCount, dest);
			}else if(opcode==Instruction::GetElementPtr){
				Handle_GEP(I, opcode, instructionCount, dest);
			}else if(I->isCast()){
				errs()<<"Cast Type \n";
				for(Instruction::op_iterator opi = (*I).op_begin(); opi != (*I).op_end(); opi++){
//					errs() << opi->get()->getType()->getTypeID() << " ";

					errs() << "\t type= "<<opi->get()->getType()->getTypeID()<< " Name = "<<opi->get()->getName() << "\n";
					source[j++] = opi->get()->getName();

				}
				nPtr = new Nodes(dest, source[0], "",opcode,instructionCount); 
				theGraph.insert(std::pair<std::string,Nodes>(dest,*nPtr));

				mapIT=theGraph.find(source[0]);
				if(mapIT!=theGraph.end()){
					mapIT->second.succ.push_back(dest);
				}
				finishTimes.insert(std::pair<std::string,int>(dest,0));// map of finish times
			}else if(I->isBinaryOp()){
				Handle_Binary(I, opcode, instructionCount, dest);
			}else if (CallInst* callInst = dyn_cast<CallInst>(&*I)) {
				Handle_Call(callInst,opcode, instructionCount, dest);
          	}
		//get a pointer to just inserted map entry
		errs () << "\n";
		instructionCount++;
		}
	  }


	 /** Record BB depths **/
	std::vector<std::string>::iterator it;
	depths["entry"] = 0;
	std::queue<std::string> myqueue;
	std::set<std::string> marked;
	myqueue.push("entry");
	marked.insert("entry");

	while(!myqueue.empty()){
		CFGit = myCFG.find(myqueue.front());
		myqueue.pop();
		//POP From queue, process it
		Block * currentBlock = &(CFGit->second);
		
		int maxD= 0;
		for(unsigned int j=0;j<currentBlock->pred.size();j++){
			std::string pred = currentBlock->pred[j];
			if(depths.find(pred)!=depths.end()){
				if(depths[pred]+1>maxD)
					maxD = depths[pred]+1;
			}
		}
		depths[currentBlock->name] = maxD;
		
		for(it = currentBlock->succ.begin();it != currentBlock->succ.end(); it++){
			if(marked.count(*it)==0){
				marked.insert(*it);
				myqueue.push(*it);
			}
		}
	}


	/** Propagate Load location **/
	for ( mapIT=theGraph.begin() ; mapIT != theGraph.end(); mapIT++ ){
		//errs()<<mapIT->first<<" \n";
		//find if operation is a load, replace uses of load temporary with load source
	
		Nodes * currNode = &(mapIT->second);
		Nodes * successor = NULL;
		
		std::map<std::string, Nodes>::iterator mapIT2;
		std::string inputName = currNode->ops[0];
		
		if(currNode->opcode==Instruction::Load){
			for(unsigned int k=0;k<currNode->succ.size();k++){//look at successors
				mapIT2 = theGraph.find(currNode->succ[k]);//find entry of successor 
				if(mapIT2!=theGraph.end()){
					successor = &(mapIT2->second);
				//	errs()<<"successor->opcode"<<successor->opcode<<"\n";
					//Replace uses in the inputs
					bool r= false;
					for(int idx=0; idx<2; idx++){
						if(successor->ops[idx]==currNode->name){
							successor->ops[idx] = inputName;
							r= true;
						}
					}

					if(r){ //propagate memDep 
						for(unsigned int j=0;j<currNode->memDep.size();j++)
							successor->memDep.push_back(currNode->memDep[j]);
					}
				}
			}
		}
	}//End load
	
	
	/** Collapse convert instructions (in forward direction) **/
	for ( mapIT=theGraph.begin() ; mapIT != theGraph.end(); mapIT++ ){
		//errs()<<mapIT->first<<" \n";
		// if key is convXXX then get input, go to each successor and change usage to input
		size_t pos = mapIT->first.find("conv");
		Nodes * currNode = &(mapIT->second);
		Nodes * successor = NULL;
		std::map<std::string, Nodes>::iterator mapIT2;
		
		if(pos!=std::string::npos){
			//found conv
			std::string inputName = currNode->ops[0];
			for(unsigned int k=0;k<currNode->succ.size();k++){//look at successors
				mapIT2 = theGraph.find(currNode->succ[k]);//find entry of successor 
				if(mapIT2!=theGraph.end()){
					successor = &(mapIT2->second);
					bool r = false;
					//Replace uses in the inputs
					for(int idx=0; idx<2; idx++){
						if(successor->ops[idx]==currNode->name){
							successor->ops[idx] = inputName;
							r= true;
						}
					}
					if(r){ //propagate memDep 
						for(unsigned int j=0;j<currNode->memDep.size();j++)
							successor->memDep.push_back(currNode->memDep[j]);
					}
				}
			}
		}
	}//End flattening
	
	/** Collapse convert to stores 
		Find stores in the store datastructure that have as input a converted value,
		and replace the value with the input of the convert instruction
		
	**/
	for ( itStores=allStores.begin() ; itStores != allStores.end(); itStores++ ){
		//errs()<<mapIT->first<<" \n";
		// if key is convXXX then get input, go to each successor and change usage to input
		size_t pos = itStores->sourceVar.find("conv");
		std::string source;
		
		if(pos==0){
			//found conv
			//Look it up in map and get the input
			mapIT = theGraph.find(itStores->sourceVar);
			if(mapIT!=theGraph.end()){
				source = mapIT->second.ops[0];
				itStores->sourceVar = source;
			}
		}
	}//End flattening


	std::set<int> UniqueOpcodes;
	std::set<int>::iterator setIT;

	errs()<<"Printing graph \n";
	for ( mapIT=theGraph.begin() ; mapIT != theGraph.end(); mapIT++ ){
	    mapIT->second.print();	
		//UniqueOpcodes.insert(mapIT->second.opcode);
	}

	for(itStores=allStores.begin(); itStores !=allStores.end(); itStores++){
		itStores->print();
	}
/*
	for( setIT=UniqueOpcodes.begin(); setIT != UniqueOpcodes.end(); setIT++){
		errs()<<"Opcode "<<*setIT<<" "<<Instruction::getOpcodeName(*setIT)<<"\n";
	}
*/

/*
	  for(CFGit = myCFG.begin(); CFGit!= myCFG.end(); CFGit++){
		CFGit->second.print();
	  }

*/
/*
	for (Function::iterator b = F.begin(), e = F.end(); b != e; ++b){
	
	  BBName = b->getName();
	  //errs()<<"\n BB name = "<<BBName<<" \n"; 
	  CFGit = myCFG.find(b->getName()); //Look up BB name and get a pointer to the data
		errs()<<"\n";
		
		
		errs()<<"\tNode"<<BBName <<"\n ";
		
		for (BasicBlock::iterator I = b->begin(), ie = b->end(); I != ie; ++I) {
			//Look up name in start times ( if it is schedulable it will have an entry)
			std::string name=I->getName();
			errs() << *I <<" ";
			timesIT= startTimes.find(name);
			if(timesIT!=startTimes.end()){
				errs()<<"(( "<<timesIT->second <<" ))";
			}
			errs()<< " \n ";
		}		
	}
*/

	/* Must run in this order */
	GEP_mappings();

	memoryLocations();
	/* Must run in this order */

	
	schedule();
/*
	for(loadTimesIT=loadTimes.begin();loadTimesIT!=loadTimes.end();loadTimesIT++){
		errs()<<loadTimesIT->first<<" : "<<loadTimesIT->second<<" \n ";
	}
	*/
/**
Print for .DOT graphviz format
**/
//	printDOT(F);
	errs()<<" Print sched \n";
	schedulePrinter(F);


/*
	for(memoryMapIT=memoryMap.begin(); memoryMapIT != memoryMap.end(); memoryMapIT++){
		errs()<<" Variable/value "<<memoryMapIT->first<<" index stored "<<memoryMapIT->second<<" \n ";
	}
*/
	/*
	for(timesIT=startTimes.begin(); timesIT != startTimes.end(); timesIT++){
		errs()<<" Operation "<<timesIT->first<<", start time "<<timesIT->second<<" \n ";
	}
	*/
      return false;
    }
  };
}

char DFG::ID = 0;
static RegisterPass<DFG> X("dfgraph", "Ilian's DFG graph", true, false);

namespace {
  // Hello2 - The second implementation with getAnalysisUsage implemented.
  struct Hello2 : public FunctionPass {
    static char ID; // Pass identification, replacement for typeid
    Hello2() : FunctionPass(ID) {}

    virtual bool runOnFunction(Function &F) {
      ++HelloCounter;
      errs() << "Hello: ";
      errs().write_escaped(F.getName()) << '\n';
      return false;
    }

    // We don't modify the program, so we preserve all analyses
    virtual void getAnalysisUsage(AnalysisUsage &AU) const {
      AU.setPreservesAll();
    }
  };
}

char Hello2::ID = 0;
static RegisterPass<Hello2>
Y("hello2", "Hello World Pass (with getAnalysisUsage implemented)");
