#include <sstream>
using namespace std;

#include "Config.h"

Config::Config(const char *configInputFile) {
	FILE *inputFile = fopen(configInputFile, "r");
	assert(inputFile != NULL);

	parseSettingsConfig(inputFile);
	fclose(inputFile);
}

Config::~Config() {
}

void Config::parseSettingsConfig(FILE *inputFile) {
	char lineBuffer[MAX_CONFIG_LINE_LEN], field[MAX_CONFIG_LINE_LEN], value[MAX_CONFIG_LINE_LEN];
	
	while (fgets(lineBuffer, MAX_CONFIG_LINE_LEN, inputFile) != NULL) {
		if (lineBuffer[0] == '#' || lineBuffer[0] == '\n')
			continue;

		assert(sscanf(lineBuffer, "%s = %s", field, value) == 2);
		strMap[field] = value;
	}
}

void Config::writeSettingsToFile(const char *configOutputFile) {
	FILE *outputFile = fopen(configOutputFile, "w");
	assert(outputFile != NULL);

	map<string, string>::const_iterator i;
	for (i = strMap.begin(); i !=strMap.end(); i++) {
		if (i->second[0] != '\0')
			fprintf(outputFile, "%s = %s\n", i->first, i->second);
	}

	fclose(outputFile);
}

void Config::writeSettingsToConsole() {
	map<string, string>::const_iterator i;
	for (i = strMap.begin(); i !=strMap.end(); i++) {
		if (i->second[0] != '\0')
			printf("%s = %s\n", i->first.c_str(), i->second.c_str());
	}
	printf("\n");
}

void Config::addStrField(const string &field, const string &value) {
	strMap[field] = value;
}

void Config::updateField(const string &field, const string &value) {
	map<string, string>::const_iterator match;

	match = strMap.find(field);
	if (match != strMap.end()) strMap[field] = value;
	else {
		printf("Unknown string field: %s\n", field);
		exit(-1);
	}
}

void Config::updateField(const string &field, const uint &value) {
	stringstream out;
	out << value;
	updateField(field, out.str());
}

string Config::getStr(const string &field) const {
	map<string, string>::const_iterator match;
	
	match = strMap.find(field);
	if (match != strMap.end()) return match->second;
	else {
		printf("Unknown string field: %s\n", field);
		exit(-1);
	}
}

uint Config::getUInt(const string &field) const {
	return (atoi(getStr(field).c_str()));
}
