// gsl_lu.cpp : main project file.

#include "stdafx.h"
#include <iostream>
#include <fstream>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_rng.h>
#include <time.h>
#include <Windows.h>


using namespace System;
using namespace std;

void WriteMatrixToFile(const char* outputFileName, gsl_matrix *matrix, const int matrixSize)
{
    ofstream outputFile(outputFileName);
    if (!outputFile.is_open())
    {
        printf("Unable to open output matrix file for writing.\n");
        exit(1);
    }
    
    outputFile << matrixSize << "\n";
    
    for (int i=0; i<matrixSize; i++)
    {
        for (int j=0; j<matrixSize; j++)
            outputFile << gsl_matrix_get(matrix, i, j) << " ";
        outputFile << "\n";
    }
    
    outputFile.close();
}

gsl_matrix *ReadArguments(const int &argc, char **argv, int &matrixSize, int &blkSize)
{
    if (argc != 3)
    {
        printf("Invalid arguments provided.\n");
        printf("Required arguments: <path to input matrix file> <path to output matrix file>");
        exit(1);
    }
    
    ifstream inputFile(argv[1]);
    if (!inputFile.is_open())
    {
        printf("Unable to open input matrix file for reading.\n");
        exit(1);
    }
    
    inputFile >> matrixSize >> blkSize;
        
	gsl_matrix *matrix = gsl_matrix_alloc (matrixSize, matrixSize);
    
	for (int i = 0; i < matrixSize; i++)
		for (int j = 0; j < matrixSize; j++)
		{	
			double entry;
			inputFile >> entry;
			gsl_matrix_set(matrix, i, j, entry);     
		}
		
    inputFile.close();
    
    return (matrix);
}


void make_random_matrix(gsl_matrix *matrix, int N)
{
	// initialize random seed:
	int i,j;
	srand ( time(NULL) );
	for (i = 0; i < N; i++)
		for (j = 0; j < N; j++)
		{
			double random = (rand() % 900 + 100) + ((rand() % 900 + 100)/1000.0);
			gsl_matrix_set(matrix, i, j, random);     
		}

	// make sure the diagonal element is larger than
	// the sum of absolute value of other numbers on the same row.
	for(i=0; i<N; i++)
	{
		double sum=0;
		for(j=0; j<N; j++)
			sum += fabs(gsl_matrix_get(matrix, i, j));
		gsl_matrix_set(matrix, i, i, sum);
	}
}

void print_matrix_to_file(char name[], gsl_matrix *matrix, int N)
{
	ofstream out_stream;
	out_stream.open(name);
	if (out_stream.fail())
	{
		printf("Could not write to file\n");
		exit(1);
	}

	int i,j;
	for(i = 0; i < N; i++)
	{
		for (j = 0; j < N; j++)
		{
			out_stream << gsl_matrix_get(matrix, i, j);
			if(j != N-1)
				out_stream << " ";
		}
		out_stream << "\n";
	}
	out_stream.close();
}

int main(int argc, char **argv)
{	
	time_t start,end;
	double run_time;
	
	int i,j;
	int blkSize=0, matrixSize=0;

	// Read matrix from file 
	// Here, blkSize is not really used. This is a non-Blocked code.
	gsl_matrix *matrix = ReadArguments(argc, argv, matrixSize, blkSize);

	// allocate the permutation
	gsl_permutation *permutation = gsl_permutation_alloc (matrixSize);
	int SIGNUM;

	// start the timer!
	time (&start);

	// do LU factorization
	gsl_linalg_LU_decomp (matrix, permutation, &SIGNUM);

	// Calculate the time
    time (&end);
	run_time = difftime (end,start);


	// print the output matrix to file
	WriteMatrixToFile(argv[2], matrix, matrixSize);

	for(i=0; i<matrixSize; i++)
		if(gsl_permutation_get(permutation,i) != i)
			printf("Error: pivoting is needed to solve this LUDecomposition. This algorithm does not pivot!\n");
	
	gsl_matrix_free (matrix);
	gsl_permutation_free(permutation);

	// Write run-time to file. The file must be called "runtime.txt" (expected by tester module)
    ofstream outputRunTime("runtime.txt");
    if (!outputRunTime.is_open())
    {
        printf("Unable to open output runtime file for writing.\n");
        exit(1);
    }
    
    outputRunTime << run_time;
	outputRunTime.close();
	
    return 0;
}
