// This file contains algorithms specific for each block type
#include "helper.h"

#include <sstream>
#include <map>

extern float* root;
extern map<float *, int> m;

string matrixIndex(float* block, uint size) {
	int row = (block - root) / size;
	int col = (block - root) % size;
	
	ostringstream convert;
	convert << "a[" << row << "][" << col << "]v";
	
	return convert.str();
}

void LUDecompType1(float *block, uint blkSize, uint size, ofstream& outputFile)
{
    for (uint ref_row=0; ref_row<blkSize-1; ref_row++)
    {
        //column normalization
        for (uint row_to_be_normalized=ref_row+1; row_to_be_normalized<blkSize; row_to_be_normalized++) {
            block[row_to_be_normalized*size+ref_row] /= block[ref_row*size+ref_row];

			outputFile << "comp " << matrixIndex(&block[row_to_be_normalized*size+ref_row],size) << m[&block[row_to_be_normalized*size+ref_row]]+1
			           << " "     << matrixIndex(&block[row_to_be_normalized*size+ref_row],size) << m[&block[row_to_be_normalized*size+ref_row]]
					   << " "     << matrixIndex(&block[ref_row*size+ref_row],size) << m[&block[ref_row*size+ref_row]] << endl;

			m[&block[row_to_be_normalized*size+ref_row]]++;
		}
        
        for (uint row=ref_row+1; row<blkSize; row++)
            for (uint column=ref_row+1; column<blkSize; column++) {
                block[row*size+column] -= (block[row*size+ref_row] * block[ref_row*size+column]);

				outputFile << "comp " << "tmp"                                         << m[NULL]
						   << " "     << matrixIndex(&block[row*size+ref_row],size)    << m[&block[row*size+ref_row]]
						   << " "     << matrixIndex(&block[ref_row*size+column],size) << m[&block[ref_row*size+column]] << endl;

				outputFile << "comp " << matrixIndex(&block[row*size+column],size)     << m[&block[row*size+column]]+1
					       << " "     << matrixIndex(&block[row*size+column],size)     << m[&block[row*size+column]]
				           << " "     << "tmp"                                         << m[NULL] << endl;

				m[NULL]++;
				m[&block[row*size+column]]++;
			}
    }
}

void LUDecompType2(float *block, float *top_block, uint blkSize, uint size, ofstream& outputFile)
{
	// for each diagonal element in top_block
    for (uint ref_top=0; ref_top<blkSize; ref_top++) {
		// for each element below it in block
        for (uint row_to_be_normalized=0; row_to_be_normalized<blkSize; row_to_be_normalized++) {               
			// normalize
            block[row_to_be_normalized*size+ref_top] /= top_block[ref_top*size+ref_top];

			outputFile << "comp " << matrixIndex(&block[row_to_be_normalized*size+ref_top],size) << m[&block[row_to_be_normalized*size+ref_top]]+1
			           << " "     << matrixIndex(&block[row_to_be_normalized*size+ref_top],size) << m[&block[row_to_be_normalized*size+ref_top]]
					   << " "     << matrixIndex(&top_block[ref_top*size+ref_top],size)          << m[&top_block[ref_top*size+ref_top]] << endl;
			
			m[&block[row_to_be_normalized*size+ref_top]]++;
		}
        
		// for each element below it in block
		for (uint row=0; row<blkSize; row++) {
			// for each column right of current diagonal element
			for(uint column=ref_top+1; column<blkSize; column++) {
                block[row*size+column] -= (block[row*size+ref_top] * top_block[ref_top*size+column]);

				outputFile << "comp " << "tmp"                                             << m[NULL]
						   << " "     << matrixIndex(&block[row*size+ref_top],size)        << m[&block[row*size+ref_top]]
						   << " "     << matrixIndex(&top_block[ref_top*size+column],size) << m[&top_block[ref_top*size+column]] << endl;

				outputFile << "comp " << matrixIndex(&block[row*size+column],size)         << m[&block[row*size+column]]+1
					       << " "     << matrixIndex(&block[row*size+column],size)         << m[&block[row*size+column]]
				           << " "     << "tmp"                                             << m[NULL] << endl;

				m[NULL]++;
				m[&block[row*size+column]]++;
			}
		}
    }
}

void LUDecompType3(float *block, float *left_block, uint blkSize, uint size, ofstream& outputFile)
{
	// for each column in left-most block
    for (uint left_ref=0; left_ref<blkSize; left_ref++) {
		// for each element below it
		for (uint row=left_ref+1; row<blkSize; row++) {
			// for each column in current block
			for (uint column=0; column<blkSize; column++) {
                block[row*size+column] -= (left_block[row*size+left_ref] * block[left_ref*size+column]);

				outputFile << "comp " << "tmp"                                             << m[NULL]
						   << " "     << matrixIndex(&left_block[row*size+left_ref],size)  << m[&left_block[row*size+left_ref]]
						   << " "     << matrixIndex(&block[left_ref*size+column],size)    << m[&block[left_ref*size+column]] << endl;

				outputFile << "comp " << matrixIndex(&block[row*size+column],size)         << m[&block[row*size+column]]+1
					       << " "     << matrixIndex(&block[row*size+column],size)         << m[&block[row*size+column]]
				           << " "     << "tmp"                                             << m[NULL] << endl;

				m[NULL]++;
				m[&block[row*size+column]]++;
			}
		}
	}
}

void LUDecompType4(float *block, float *top_block, float *left_block, uint blkSize, uint size, ofstream& outputFile)
{
	// for each column in left-most block
	for (uint left_ref=0; left_ref<blkSize; left_ref++) {
		// for each element in current block
		for (uint row=0; row<blkSize; row++) {
			// for each column in top-most block
			for (uint column=0; column<blkSize; column++) {
				block[row*size+column] -= (left_block[row*size+left_ref] * top_block[left_ref*size+column]);

				outputFile << "comp " << "tmp"                                              << m[NULL]
						   << " "     << matrixIndex(&left_block[row*size+left_ref],size)   << m[&left_block[row*size+left_ref]]
						   << " "     << matrixIndex(&top_block[left_ref*size+column],size) << m[&top_block[left_ref*size+column]] << endl;

				outputFile << "comp " << matrixIndex(&block[row*size+column],size)          << m[&block[row*size+column]]+1
					       << " "     << matrixIndex(&block[row*size+column],size)          << m[&block[row*size+column]]
				           << " "     << "tmp"                                              << m[NULL] << endl;

				m[NULL]++;
				m[&block[row*size+column]]++;
			}
		}
	}
}
