#include <iostream>
#include <sstream>
using namespace std;

#include "Node.h"

//actual definition
map<string,uint> Node::dataIDMap;
#ifdef DEBUG
map<uint,string> Node::IDtoName;
#endif

//static definition of getDataID method
uint Node::getDataID(string name) {
	map<string,uint>::iterator it = dataIDMap.find(name);

	if (it == dataIDMap.end()) {
		uint newID = dataIDMap.size();
		dataIDMap[name] = newID;
		#ifdef DEBUG
			IDtoName[newID] = name;
		#endif
		return newID;
	}

	return dataIDMap[name];
}

Node::Node() {
	arrivedPackets = new queue<Packet>;
	incomingPackets = new queue<Packet>;
}

Node::~Node() {
	outgoingPackets.clear();
	routingTable.clear();
	neighbours.clear();

	delete arrivedPackets;
	delete incomingPackets;

	outputFile->close();
	delete outputFile;
}

void Node::setupNode(Config* config, uint nodeId) {
	ostringstream outputFileName;
	this->config = config;
	this->nodeId = nodeId;
	outputFileName << config->getStr("OutputFilePrefix") << nodeId << config->getStr("OutputFileSuffix");
	outputFile = new ofstream(outputFileName.str().c_str());

	if (!outputFile->is_open()) {
		cout << "Output file " << outputFileName.str().c_str() << "could not be opened.\n";
		exit(-1);
	}
}

uint Node::getNodeId() {
	return (nodeId);
}

void Node::setNodeId(uint nodeId) {
	this->nodeId = nodeId;
}

bool Node::isNeighbour(Node* key) {
	for (uint i=0; i < neighbours.size(); i++) {
		if (neighbours[i] == key)
			return (true);
	}
	return (false);
}

void Node::addNeighbour(Node* newEntry) {
	neighbours.push_back(newEntry);
}

bool Node::isEntryInRoutingTable(uint destNodeId) {
	return (routingTable.find(destNodeId) != routingTable.end());
}

void Node::addRoutingEntry(uint destNodeId, Node *nextHopAddr) {
	routingTable[destNodeId] = nextHopAddr;
}
