#ifndef NODE_H
	#define NODE_H

	#include <vector>
	#include <queue>
	#include <set>
    #include <fstream>
	#include "../Common.h"
	#include "../Config/Config.h"

	enum PacketType {
		PKT_SEND,
		PKT_REQUEST
	};

	struct Packet {
		PacketType t;
		uint destNodeID;
		uint dataID;
		uint returnNodeID;		
	};

	struct Channel {
		vector<Packet> queue;
		vector<uint> latencyRemaining;
	};

	class Node {
		private:
			static map<string,uint> dataIDMap; //external declaration			

		protected:
			#ifdef DEBUG
				static map<uint,string> IDtoName;
			#endif

			Config *config;
			uint nodeId;

			vector<Node*> neighbours;
			map<uint,Node*> routingTable;
			queue<Packet>* incomingPackets;
			queue<Packet>* arrivedPackets;
			vector<Packet> outgoingPackets;
			
			ofstream* outputFile;

			// Return the data ID of a given string, if the name is not associated with an ID yet, it will assign one
			static uint getDataID(string name);
		
		public:
			// Constructors
			Node();
			~Node();

			uint getNodeId();
			void setNodeId(uint nodeId);

			bool isNeighbour(Node* key);
			void addNeighbour(Node* newEntry);

			bool isEntryInRoutingTable(uint destNodeId);
			void addRoutingEntry(uint destNodeId, Node *nextHopAddr);

			void setupNode(Config* config, uint nodeId);

			// Setup before starting the step again
			virtual void preStep() = 0;

			// Tell the node to do 1 cycle of work
			virtual void step() = 0;

			// The node attempt to receive a packet
			// The node should rely the packet towards destination node if the node does not have the same ID as destNodeId
			// Return true if successfully received
			// Return False if the node cannot receive the word for whatever reason
			virtual bool receive(Packet p) = 0;

			virtual bool send(set<Node*>& usedLinks, Packet p) = 0;

			// Add an instruction for the node to process
			virtual void addInstruction(string insn) = 0;

			// Return if the node has finished all its instruction or not
			virtual bool isCompleted() = 0;

			// Reset counter
			virtual void resetCounters() = 0;

			// Print Stat
			virtual void printStat() = 0;
	};
#endif
