#ifndef SIMPLENODE_H
	#define SIMPLENODE_H

	#include "Node.h"

	enum SimpleNodeInsnType {
		SNIT_REQUEST,
		SNIT_COMPUTE,
		SNIT_SEND,
		SNIT_WAIT
	};

	struct SimpleNodeInsn {
		SimpleNodeInsnType t;
		uint output1;
		uint input1, input2;
	};

	class SimpleNode : public Node {
		private:
			// SimpleNode Parameters
			uint InstructionWindowSize;
			uint InstructionsPerCycle;
			uint NumberIOPorts;
			uint NumberBRAMWritePerCycle;
			uint NumberBRAMReadPerCycle;

			// Counters
			uint requSent;
			uint requReceived;
			uint dataReceived;
			uint dataSent;
			uint idleCycle;
			uint compCycle;
			uint sendCycle; 

			// Instance Data
			set<uint> availableData;
			set<uint> ioports;
			set<uint> tobeWritten;
			
			// Instructions for the instance node
			deque<SimpleNodeInsn> instructions;

		protected:
			bool send(set<Node*>& usedLinks, Packet p);

		public:
			// Constructors
			SimpleNode();			
			~SimpleNode();

			void preStep();
			void step();

			void setupNode(Config* c, uint nodeId);

			bool receive(Packet p);			

			// Add an instruction for the node to process
			void addInstruction(string insn);

			// Return if the node has finished all its instruction or not
			bool isCompleted();

			// Reset counters
			void resetCounters();

			// print stats
			void printStat();
	};
#endif
