#include <fstream>
#include <math.h>
#include <assert.h>
#include <iostream>
using namespace std;

#include "GeneratePeers.h"

#ifdef TOFILE
	void GeneratePeers(char *fileName, uint size, uint &numPeers)
#else
	uint **GeneratePeers(uint size, uint &numPeers)
#endif
{
	double boxSize = sqrt(size);
	assert(boxSize == floor(boxSize));

	uint puzzleCells = size*size;
	numPeers = (size-1)*2 + (uint) ((boxSize-1)*(boxSize-1));

	#ifdef TOFILE
		ofstream outputFile(fileName);
		assert(outputFile.is_open());

		outputFile << "static const uint peers[" << puzzleCells << "][" << numPeers << "] = {" << endl;
	#else
		uint **peers = new uint*[puzzleCells];
		for (uint i=0; i < puzzleCells; i++) {
			peers[i] = new uint[numPeers];
		}
	#endif

	for (uint row=0; row < size; row++) {
		for (uint col=0; col < size; col++) {
			#ifdef TOFILE
				bool firstOne = true;			
				outputFile << " { ";
			#else
				uint index = 0;
			#endif

			//same row
			for (uint i=0; i < size; i++) {
				uint add = row*size + i;
				if (add != row*size + col) {
					#ifdef TOFILE
						if (firstOne) {						
							outputFile << add;						
							firstOne = false;
						}
						else
							outputFile << ", " << add;
					#else
						peers[row*size+col][index] = add;
						index++;
					#endif
				}
			}

			//same column
			for (uint i=0; i < size; i++) {
				uint add = i*size + col;
				if (add != row*size + col) {
					#ifdef TOFILE
						outputFile << ", " << add;
					#else
						peers[row*size+col][index] = add;
						index++;
					#endif
				}
			}

			//same box
			uint rowBox = (uint) floor(row / boxSize);
			uint colBox = (uint) floor(col / boxSize);
			
			uint currBox = (uint) (rowBox*boxSize*size + colBox*boxSize);

			//currRowBox: Current Row wrt to a box. For a 9x9 puzzle, currRowBox = 0 to 2
			//currBox: Points to the first item for a given box
			//Iterates over the elements in a box (3x3 box for a 9x9 puzzle)

			for (uint boxRow=0; boxRow < boxSize; boxRow++) {
				for (uint boxCol=0; boxCol < boxSize; boxCol++) {
					uint add = currBox + boxRow*size + boxCol;

					//dont add itself, same column or same row
					if (add != row*size + col && (add-col)%size != 0 && add-row*size >= size) {
						#ifdef TOFILE
							outputFile << ", " << add;
						#else
							peers[row*size+col][index] = add;
							index++;
						#endif
					}
				}
			}
		
			#ifdef TOFILE
				outputFile << " }";
				if (row != size-1 || col != size-1)
					outputFile << "," << endl;
			#endif
		}
	}
	
	#ifdef TOFILE
		outputFile << endl << "};" << endl;
		outputFile.close();
	#else		
		return peers;
	#endif
}
