#ifndef SERIAL_SOLVER_H
	#define SERIAL_SOLVER_H

	#include "../../Common.h"
	#include <fstream>
	#include <map>

	class SudokuSerialSolver {
		/* STRUCTURES */
		struct Cell {
			uint possibilities; //For 9x9 puzzle, 0b000101100 means 3, 4 and 6 are possible candidates
			uint numPossible;   //For above example, this would be 3
		};

		/* STATIC MEMBERS */
		private:
			static map<uint, uint> ValueLookup;
			static void populateLookupTable(uint size);

			static uint numPeers, **peers;

			static map<Cell*, uint> cellMap;
			static string cellOut(uint row, uint col);

		/* INSTANCE MEMBERS */
		public:
			SudokuSerialSolver(char *inPuzzleFile, char *outInsnFilePath);			
			~SudokuSerialSolver();

			void solve();
			void printPuzzleToScreen();
			void printPuzzleToFile(char* fileName);
			bool verify(uint& errRow, uint& errCol);
			bool verifyUnsolved(uint& errRow, uint& errCol);

		private:
			uint size; //For 9x9, size = 9
			uint numSingletons; //Number of cells with 1 possibility
			Cell *cells;        //stores the puzzle

			ofstream *outInsnFile;

			//solve the puzzle using constraint propagation
			//Refer to http://norvig.com/sudoku.html
			void doConstraintPropagation();

			uint *createPuzzleGrid();
	};
#endif
