#include <iostream>
#include <math.h>
using namespace std;

#include "Validator.h"

/**
 * Check if a solved puzzle is valid.
 * puzzle: 1D puzzle array.
 * size: Size of the array.
 * errRow: If validation fails, item (errRow, errCol) is the cause of it.
 * errCol: If validation fails, item (errRow, errCol) is the cause of it.
 * Returns true on validation success. Otherwise false.
 */
bool ValidateSolvedSudoku(uint *puzzle, uint size, uint& errRow, uint& errCol) {
	Debug << "Verifying Rows..." << endl;

	uint mask, maskBit;
	for (uint row=0; row < size; row++) {
		//mask: 0x000010110 => For a given row, #s 2, 3 and 5 have been found
		mask = 0;
		for (uint col=0; col < size; col++) {		
			if (puzzle[row*size+col] == 0) {
				errRow = row;
				errCol = col;
				return (false);
			}

			//maskBit: Current number - 0x000010000 => 5
			maskBit = 1 << (puzzle[row*size+col] - 1);

			Debug << "Comparing " << puzzle[row*size+col]
			      << " Mask " << mask << " MaskBit " << maskBit
			      << " Ans " << (maskBit & mask) << endl;

			//if the number didn't occur yet
			if ((maskBit & mask) == 0)
				mask += maskBit;

			//otherwise, validation failure
			else {
				errRow = row;
				errCol = col;
				return false;
			}
		}
	}

	Debug << "Verified! Verifying Columns..." << endl;

	for (uint col=0; col < size; col++) {
		mask = 0;
		for (uint row=0; row < size; row++) {
			maskBit = 1 << (puzzle[row*size+col] - 1);

			Debug << "Comparing " << puzzle[row*size+col]
			      << " Mask " << mask << " MaskBit " << maskBit
			      << " Ans " << (maskBit & mask) << endl;

			if ((maskBit & mask) == 0)
				mask += maskBit;
			else {
				errRow = row;
				errCol = col;
				return false;
			}
		}
	}

	Debug << "Verified! Verifying boxes..." << endl;

	//currRowBox: Current Row wrt to a box. For a 9x9 puzzle, currRowBox = 0 to 2
	//currBox: Points to the first item for a given box
	uint *currRowBox, *currBox;

	//For a 9x9 puzzle, the box size is 3
	uint boxSize = (uint) sqrt(size);

	//The first two for-loops jumps to a box (from 0 to 8 for a 9x9 puzzle)
	for (uint rowBox=0; rowBox < boxSize; rowBox++) {
		currRowBox = puzzle + rowBox*boxSize*size;
		for (uint colBox=0; colBox < boxSize; colBox++) {
			currBox = currRowBox + colBox*boxSize;
			mask = 0;

			//Iterates over the elements in a box (3x3 box for a 9x9 puzzle)
			for (uint boxRow=0; boxRow < boxSize; boxRow++) {
				for (uint boxCol=0; boxCol < boxSize; boxCol++) {
					maskBit = 1 << (currBox[boxRow*size + boxCol] - 1);

					Debug << "Comparing " << currBox[boxRow*size + boxCol]
						  << " Mask " << mask << " MaskBit " << maskBit
						  << " Ans " << (maskBit & mask) << endl;

					if ((maskBit & mask) == 0)
						mask += maskBit;
					else {
						errRow = rowBox*boxSize + boxRow;
						errCol = colBox*boxSize + boxCol;
						return false;
					}
				}
			}
		}
	}

	return true;
}

/**
 * Check if an unsolved puzzle is valid.
 * puzzle: 1D puzzle array.
 * size: Size of the array.
 * errRow: If validation fails, item (errRow, errCol) is the cause of it.
 * errCol: If validation fails, item (errRow, errCol) is the cause of it.
 * Returns true on validation success. Otherwise false.
 */
bool ValidateUnsolvedSudoku(uint *puzzle, uint size, uint& errRow, uint& errCol) {
	Debug << "Verifying Rows..." << endl;

	uint mask, maskBit;
	for (uint row=0; row < size; row++) {
		//mask: 0x000010110 => For a given row, #s 2, 3 and 5 have been found
		mask = 0;
		for (uint col=0; col < size; col++) {
			if (puzzle[row*size+col] == 0) {
				continue;
			}

			//maskBit: Current number - 0x000010000 => 5
			maskBit = 1 << (puzzle[row*size+col] - 1);

			Debug << "Comparing " << puzzle[row*size+col]
			      << " Mask " << mask << " MaskBit " << maskBit
			      << " Ans " << (maskBit & mask) << endl;

			//if the number didn't occur yet
			if ((maskBit & mask) == 0)
				mask += maskBit;

			//otherwise, validation failure
			else {
				cout << "rowerror" << endl;
				errRow = row;
				errCol = col;
				return false;
			}
		}
	}

	Debug << "Verified! Verifying Columns..." << endl;

	for (uint col=0; col < size; col++) {
		mask = 0;
		for (uint row=0; row < size; row++) {
			if (puzzle[row*size+col] == 0) {
				continue;
			}
			maskBit = 1 << (puzzle[row*size+col] - 1);

			Debug << "Comparing " << puzzle[row*size+col]
			      << " Mask " << mask << " MaskBit " << maskBit
			      << " Ans " << (maskBit & mask) << endl;

			if ((maskBit & mask) == 0)
				mask += maskBit;
			else {
				cout << "colerror" << endl;
				errRow = row;
				errCol = col;
				return false;
			}
		}
	}

	Debug << "Verified! Verifying boxes..." << endl;

	//currRowBox: Current Row wrt to a box. For a 9x9 puzzle, currRowBox = 0 to 2
	//currBox: Points to the first item for a given box
	uint *currRowBox, *currBox;

	//For a 9x9 puzzle, the box size is 3
	uint boxSize = (uint) sqrt(size);

	//The first two for-loops jumps to a box (from 0 to 8 for a 9x9 puzzle)
	for (uint rowBox=0; rowBox < boxSize; rowBox++) {
		currRowBox = puzzle + rowBox*boxSize*size;
		for (uint colBox=0; colBox < boxSize; colBox++) {
			currBox = currRowBox + colBox*boxSize;
			mask = 0;

			//Iterates over the elements in a box (3x3 box for a 9x9 puzzle)
			for (uint boxRow=0; boxRow < boxSize; boxRow++) {
				for (uint boxCol=0; boxCol < boxSize; boxCol++) {
					if (currBox[boxRow*size + boxCol] == 0) {
						continue;
					}
					maskBit = 1 << (currBox[boxRow*size + boxCol] - 1);

					Debug << "Comparing " << currBox[boxRow*size + boxCol]
					      << " Mask " << mask << " MaskBit " << maskBit
					      << " Ans " << (maskBit & mask) << endl;

					if ((maskBit & mask) == 0)
						mask += maskBit;
					else {
						cout << "boxerror" << endl;
						errRow = rowBox*boxSize + boxRow;
						errCol = colBox*boxSize + boxCol;
						return false;
					}
				}
			}
		}
	}

	return true;
}
